CHUNK_SIZE = 5 * 1024 * 1024  # 5 MB
CHUNK_SIZE_LARGE = CHUNK_SIZE * 10  # 50 MB
DEFAULT_MODE = "rb"
DEFAULT_WRITE_MODE = "wb"

SCHEME_FILE = "file"
SCHEME_S3 = "s3"
SCHEME_REDIS = "redis"
SCHEME_MEMORY = "memory"
